<?php

namespace FiloBlu\Flow\Model\Actions\From;

use Exception;
use Magento\Catalog\Api\ProductRepositoryInterface;
use Magento\Catalog\Model\Product;
use Magento\Catalog\Model\ResourceModel\Category\CollectionFactory;

/**
 * Class Addtocategory
 * @package FiloBlu\Flow\Model\Actions\From
 */
class Addtocategory
{
    /**
     * @var array
     */
    protected $categoryLookup = [];

    /**
     * @var CollectionFactory
     */
    protected $collectionFactory;

    /**
     * @var array
     */
    protected $excludedCategories = [];

    /**
     * @var ProductRepositoryInterface
     */
    protected $productRepository;

    /**
     * @var \FiloBlu\Flow\Helper\Product
     */
    protected $productHelper;

    /**
     * TestAddToCategory constructor.
     * @param ProductRepositoryInterface $productRepository
     * @param CollectionFactory $collectionFactory
     * @param \FiloBlu\Flow\Helper\Product $productHelper
     */
    public function __construct(
        ProductRepositoryInterface $productRepository,
        CollectionFactory $collectionFactory,
        \FiloBlu\Flow\Helper\Product $productHelper
    ) {
        $this->collectionFactory = $collectionFactory;
        $this->productRepository = $productRepository;
        $this->productHelper = $productHelper;
    }

    /**
     * @param string $categoriesString single number (58) single string (parka) path of strings (woman/jacket/parka) path of number (2/5/9) separated by |
     * @param Product $product
     * @param $storeId
     * @throws Exception
     */
    public function process($product, $categoriesString, $storeId)
    {
        $this->init();

        $categories = explode('|', $categoriesString);
        $categoryIds = [];

        foreach ($categories as $categoryGroup) {
            $pieces = explode('/', $categoryGroup);
            foreach ($pieces as $category) {
                $categoryIds = array_merge($this->lookupCategory($category), $categoryIds);
            }
        }

        if (empty($categoryIds)) {
            return;
        }

        $existingCategories = $product->getCategoryIds();
        $uniqueCategories = array_unique(array_merge($categoryIds, $existingCategories));
        $categoriesToSet = array_diff($uniqueCategories, $this->excludedCategories);

        if (empty($categoriesToSet)) {
            return;
        }

        $product->setCategoryIds($categoriesToSet);
        $this->productRepository->save($product);
    }

    /**
     *
     */
    public function init()
    {
        if (!empty($this->categoryLookup)) {
            return;
        }

        $collection = $this->collectionFactory->create()
            ->setStore(0)
            ->addNameToResult();

        foreach ($collection as $category) {
            $structure = explode('/', $category->getPath());

            if (in_array($category->getLevel(), [0, 1], false)) {
                $this->excludedCategories[] = $category->getId();
            }

            // Removing the category ID 1, I don't need that
            if (($key = array_search('1', $structure, false)) !== false) {
                unset($structure[$key]);
            }

            if (count($structure) === 0) {
                continue;
            }

            $leaf_id = end($structure);
            $this->categoryLookup[$leaf_id] = $structure;
        }
    }

    /**
     * @param $category
     * @return mixed
     */
    public function lookupCategory($category)
    {
        if (is_numeric($category)) {
            if (isset($this->categoryLookup[$category])) {
                if ($this->productHelper->assignFullTree()) {
                    return $this->categoryLookup[$category];
                }
                return [$category];
            }

            return [];
        }
    }
}
