<?php

namespace FiloBlu\Flow\Model\From\Entity;

use Exception;
use FiloBlu\Flow\Model\Inboundflow;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DB\Adapter\AdapterInterface;
use Magento\Framework\Registry;
use Magento\Store\Model\ScopeInterface;

/**
 * Class Bulk
 * @package FiloBlu\Flow\Model\From\Entity
 */
class Bulk
{
    /**
     * @var AdapterInterface
     */
    protected $connection;
    /**
     * @var int
     */
    protected $_productEntityTypeId = 4;
    /**
     * @var array
     */
    protected $_excludeFromBulk = ['sku'];
    /**
     * @var ProductMetadataInterface
     */
    protected $_productMetadata;
    /**
     * @var Registry
     */
    protected $_registry;

    /**
     * @var ScopeConfigInterface
     */
    protected $_scopeConfig;

    /**
     * @var ResourceConnection
     */
    protected $resourceConnection;

    /**
     * Bulk constructor.
     * @param ResourceConnection $connection
     * @param ProductMetadataInterface $productMetadata
     * @param Registry $registry
     * @param ScopeConfigInterface $scopeConfig
     */
    public function __construct(
        ResourceConnection $connection,
        ProductMetadataInterface $productMetadata,
        Registry $registry,
        ScopeConfigInterface $scopeConfig
    ) {
        $this->resourceConnection = $connection;
        $this->_productMetadata = $productMetadata;
        $this->_registry = $registry;
        $this->_scopeConfig = $scopeConfig;
    }

    /**
     * @param $file
     * @throws Exception
     */
    public function _insertMissingAttributeOptionValues($file)
    {

        //selects values of dropdown attributes with no option yet
        $query = "SELECT eavalues.value as value, eavalues.attribute_id FROM (SELECT e.v as value, ea.attribute_id
			    	FROM `flow_from_eavs` e
			    	JOIN `eav_attribute` ea
				    	ON e.`a` = ea.`attribute_code`
				    	AND ea.`entity_type_id` = {$this->_productEntityTypeId}
				    	AND ea.`frontend_input` = 'select'
				    	AND (e.s=0 OR e.s is null)
				    	AND e.a NOT REGEXP '^_'
				    	AND e.`meta_processed` = 0
				    	AND e.meta_file = {$file->getId()}
				    	GROUP BY attribute_id,value) AS eavalues
			    	LEFT JOIN eav_attribute_option_value option_value
			    		ON option_value.value=eavalues.value
			    	LEFT JOIN eav_attribute_option eaoption
			    		ON eaoption.attribute_id=eavalues.attribute_id
			    		AND option_value.option_id=eaoption.option_id
			    	WHERE option_value.value_id IS null
			    	GROUP BY eavalues.attribute_id,eavalues.value;";

        $rows = $this->getConnection()->fetchAll($query);

        $eavAttributeTable = $this->resourceConnection->getTableName('eav_attribute_option');
        $eavAttributeValueTable = $this->resourceConnection->getTableName('eav_attribute_option_value');

        foreach ($rows as $row) {
            try {
                $this->getConnection()->insert($eavAttributeTable, ['attribute_id' => $row['attribute_id']]);
                $this->getConnection()->query("INSERT INTO `{$eavAttributeValueTable}`(`option_id`,`store_id`,`value`) VALUES(LAST_INSERT_ID(),0," . $this->getConnection()->quote($row['value']) . ");");
            } catch (Exception $e) {
                throw new Exception(__METHOD__ . ':' . __LINE__ . ': Rollback happened.' . $e->getMessage());
            }
        }
    }

    /**
     * @return AdapterInterface
     */
    protected function getConnection()
    {
        if ($this->connection === null) {
            $this->connection = $this->resourceConnection->getConnection();
        }

        return $this->connection;
    }

    /**
     * @param $file
     * @throws Exception
     */
    public function _setMetaStatusToNotFoundProducts($file)
    {
        $query = "UPDATE `flow_from_eavs` e
					LEFT JOIN catalog_product_entity as products on products.sku=e.e
					SET e.`meta_processed` = 3
					WHERE e.`meta_processed` = 0
					AND e.a NOT REGEXP '^_'
					AND (e.`meta_ref_id` != 2)
					AND products.sku is null
					AND e.meta_file = {$file->getId()};";
        try {
            $this->getConnection()->query($query);
        } catch (Exception $e) {
            throw new Exception(__METHOD__ . ':' . __LINE__ . ': Rollback happened.' . $e->getMessage());
        }
    }

    /**
     * @param $file
     * @param int $limit
     * @param int $offset
     * @return array|bool
     */
    public function _getMetaAttributesforProducts($file, $limit = 0, $offset = 0)
    {
        if ($limit > 0) {
            $limit_query = " LIMIT $limit OFFSET $offset";
        }

        $query = "SELECT cpe.`{$this->getEntityField()}`, e.`v`, IF(cs.`store_id` IS NULL, 0, cs.`store_id`) store_id, ea.`attribute_id`, ea.`backend_type`, ea.`frontend_input`, e.`meta_id`, e.`meta_file` FROM `flow_from_eavs` e
                                JOIN `catalog_product_entity` cpe
                                        ON cpe.`sku` = e.`e`
								        AND e.a NOT REGEXP '^_'
		                                AND e.`meta_processed` = 0
		                                AND meta_file = {$file->getId()}
                                JOIN `eav_attribute` ea
                                        ON e.`a` = ea.`attribute_code`
                                        AND ea.`entity_type_id` = {$this->_productEntityTypeId}
                                        AND
                                            ( (ea.`frontend_input` != 'select' AND ea.`frontend_input` != 'multiselect')
                                                OR
                                               e.a='visibility' 
                                                OR
                                               e.a='country_of_manufacture'
                                            )
                                        AND ea.attribute_code not in ('" . implode("','", $this->_excludeFromBulk) . "')
                                LEFT JOIN `store` cs ON cs.`code` = e.`s`";

        $dropdownQuery = "SELECT cpe.`{$this->getEntityField()}`, option_value.option_id as v, 0 store_id, ea.`attribute_id`, ea.`backend_type`, ea.`frontend_input`, e.`meta_id`, e.`meta_file`
        						FROM `flow_from_eavs` e
                                JOIN `catalog_product_entity` cpe
                                        ON cpe.`sku` = e.`e`
        								AND e.a NOT REGEXP '^_'
        								AND e.a!='visibility'
        								AND e.a!='country_of_manufacture'
        								AND (e.s=0 OR e.s IS NULL)
                                		AND e.`meta_processed` = 0
		                                AND meta_file = {$file->getId()}
                                JOIN `eav_attribute` ea
                                        ON e.`a` = ea.`attribute_code`
                                        AND ea.`entity_type_id` = {$this->_productEntityTypeId}
                                        AND ea.`frontend_input` = 'select'
                                        AND ea.attribute_code not in ('" . implode("','", $this->_excludeFromBulk) . "')
                                JOIN eav_attribute_option eaoption
					    			ON eaoption.attribute_id=ea.attribute_id
					    		JOIN eav_attribute_option_value option_value
					    			ON option_value.value=e.v
					    			AND option_value.option_id=eaoption.option_id
					    			AND option_value.store_id=0 $limit_query;";

        $query .= ' UNION ' . $dropdownQuery;

        $rows = $this->getConnection()->fetchAll($query);

        $toProcess = [];

        foreach ($rows as $row) {
            if (!array_key_exists($row['backend_type'], $toProcess) || !is_array($toProcess[$row['backend_type']])) {
                $toProcess[$row['backend_type']] = [];
            }
            $toProcess[$row['backend_type']][] = $row;
        }

        return empty($toProcess) ? false : $toProcess;
    }

    /**
     * TODO : create an helper shared between classes
     * @return string
     */
    protected function getEntityField()
    {
        $edition = $this->_productMetadata->getEdition();
        return ($edition === 'Enterprise' || $edition === 'B2B') ? 'row_id' : 'entity_id';
    }

    /**
     * @param Inboundflow $inboundflow
     */
    public function bulkCopyAttributesToConfigurable(Inboundflow $inboundflow)
    {
        $lookupAttributeTable = [

            'datetime' => $this->getConnection()->getTableName('catalog_product_entity_datetime'),
            'int' => $this->getConnection()->getTableName('catalog_product_entity_int'),
            'text' => $this->getConnection()->getTableName('catalog_product_entity_text'),
            'varchar' => $this->getConnection()->getTableName('catalog_product_entity_varchar'),
            'decimal' => $this->getConnection()->getTableName('catalog_product_entity_decimal')
        ];

        $excludedAttributes = $this->_excludeFromBulk;

        $attributesToCopy = $this->getAttributesToCopy(); //$this->_registry->registry('attribute_configurable_copy');

        $allowedAttributesToCopy = array_diff($attributesToCopy, $excludedAttributes);

        if (empty($allowedAttributesToCopy)) {
            return;
        }

        $sqlAttributes = [];

        foreach ($allowedAttributesToCopy as $allowedAttribute) {
            $sqlAttributes[] = $this->getConnection()->quote($allowedAttribute);
        }

        $attributesSet = join(',', $sqlAttributes);

        $safeAttributes = $this->getAttributesSafeToCopy($inboundflow, $attributesSet);

        /* SELECT PARENT - CHILD */

        $sql = "SELECT
	cpe.row_id as simple_row_id,
	sl.product_id as simple_entity_id,
	sl.parent_id as configurable_row_id
FROM
	catalog_product_super_link as sl
	LEFT JOIN catalog_product_entity as cpe on cpe.entity_id = sl.product_id
WHERE
	sl.product_id in (
        SELECT
			c.entity_id
		FROM
			flow_from_eavs AS f
		LEFT JOIN catalog_product_entity as c ON
			c.sku = f.e
            AND c.type_id = 'simple'
            AND f.a IN  ({$attributesSet})
        WHERE f.meta_file = ? AND f.meta_processed = 1
	)
GROUP BY
	sl.parent_id";
        $results = $this->getConnection()->fetchAll($sql, [$inboundflow->getId()]);

        foreach ($results as $entry) {
            /*
            simple_row_id,
            simple_entity_id,
            configurable_row_id
            */

            $this->copyAttrybutesToConfigurableByRowId($safeAttributes, $lookupAttributeTable, $entry['configurable_row_id'], $entry['simple_row_id']);
        }
    }

    /**
     * @return array
     */
    protected function getAttributesToCopy()
    {
        $storeScope = ScopeInterface::SCOPE_WEBSITE;

        $attributesToCopy = $this->_scopeConfig->getValue(\FiloBlu\Flow\Helper\Product::CONFIG_FLOW_ATTRIBUTE_PATH, $storeScope);

        if (!empty($attributesToCopy)) {
            return explode(',', $attributesToCopy);
        }

        return [];
    }

    /**
     * @param Inboundflow $inboundflow
     * @param $attributeSet quoted string like "'size','description','shorted_description'"
     * @return mixed
     */
    protected function getAttributesSafeToCopy(Inboundflow $inboundflow, $attributeSet)
    {
        $superAttributesTable = $this->getConnection()->getTableName('catalog_product_super_attribute');
        $eavAttributeTable = $this->getConnection()->getTableName('eav_attribute');
        $superLinkTable = $this->getConnection()->getTableName('catalog_product_super_link');

        return $this->getConnection()->fetchAll("SELECT
            `attribute_id`,
            `attribute_code`,
            `backend_type`
        FROM
            {$eavAttributeTable}
        WHERE
            `entity_type_id` = 4
            AND `attribute_code` IN ({$attributeSet})
        AND `attribute_id` NOT IN (
            SELECT
                    `attribute_id`
                FROM
                    {$superAttributesTable}
                WHERE
                    product_id IN (
            SELECT
                            sl.parent_id
                        FROM
                            {$superLinkTable} as sl
                        LEFT JOIN catalog_product_entity as cpe on
                            cpe.entity_id = sl.product_id
                        WHERE
                            sl.product_id in (
            SELECT
                                    c.entity_id
                                FROM
                                    flow_from_eavs AS f
                                LEFT JOIN catalog_product_entity as c ON
                                    c.sku = f.e
                                    AND c.type_id = 'simple'
                                    AND f.a IN ({$attributeSet})
                                WHERE
                                    f.meta_file = ?
                                    AND f.meta_processed = ?
                            )
                        GROUP BY
                            sl.parent_id
                    )
            )", [$inboundflow->getId(), Inboundflow::STATUS_PROCESSED]);
    }

    /**
     * @param array $attributesToCopy array of array  containing [ 'attribute_id' => ..., 'attribute_code' => ...., 'backend_type' => ....]
     * @param array $lookupAttributeTable
     * @param $configurableRowId
     * @param $simpleRowId
     */
    protected function copyAttrybutesToConfigurableByRowId(array $attributesToCopy, array $lookupAttributeTable, $configurableRowId, $simpleRowId)
    {
        /* TODO: to optimize -> group attributes by backend_type and perform a mass update */
        foreach ($attributesToCopy as $attributeInfo) {
            $backendType = $attributeInfo['backend_type'];

            if ($backendType === 'static') {
                // TODO :
                continue;
            }

            $table = $lookupAttributeTable[$backendType];

            $sql = "INSERT INTO {$table} ( attribute_id, store_id, `value`, row_id) SELECT `attribute_id`, `store_id`, `value`, {$configurableRowId} AS row_id FROM {$table} WHERE row_id = {$simpleRowId} AND attribute_id = {$attributeInfo['attribute_id']} ON DUPLICATE KEY UPDATE `value` = VALUES(`value`)";

            $this->getConnection()->query($sql);
        }
    }
}
